#
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Mon Feb 27 2012
#



proc oprs_service_va_doc { service } {
    if {[llength [$service parameters in inout]] == 0} {
	return ""
    } else {  
	set res ""
	foreach  p [$service parameters in inout] {
	    append res [oprs_type_va_doc  [$p type] $p [$p name]]
	}
	return $res
    }
}


proc oprs_type_va_doc { type param string } {
    switch -glob -- [$type kind] {
	typedef - {* member} {
	    return [oprs_type_va_doc [$type type] $param $string]
	}
	string - array - sequence - "unsigned long" - long - float - double - int - char - enum {
	    return " ($string <[mangle $type]>)"
	}
	struct {
	    set res ""
	    foreach e [$type members] {
		append res [oprs_type_va_doc [$e type] $param $string.[$e name]]
	    }
	    return $res
	}
	default {
	    puts stderr "unexpected type: [$type kind]" 
	}
    }
}

proc oprs_service_nva { service } {
    if {[llength [$service parameters in inout]] == 0} {
	return 0
    } else {  
	set res 0
	foreach  p [$service parameters in inout] {
	    set res [expr $res + [oprs_type_nva  [$p type] $p [$p name]]]
	}
	return $res
    }
}


proc oprs_type_nva { type param string } {
    switch -glob -- [$type kind] {
	typedef - {* member} {
	    puts stderr member
	    return [oprs_type_nva [$type type] $param $string]
	}
	string - array - sequence - "unsigned long" - long - float - double - int - char - enum {
	 	    return 1
	}
	struct {
	    set res 0
	    foreach e [$type members] {
		set res [expr $res + [oprs_type_nva [$e type] $param $string.[$e name]]]
	    }
	    return $res
	}
	default {
	    puts stderr "unexpected type: [$type kind]" 
	}
    }
}

proc oprs_service_va { service } {
    if {[llength [$service parameters in inout]] == 0} {
	return ""
    } else {  
	set res ""
	foreach  p [$service parameters in inout] {
	    append res [oprs_type_va  [$p type] $p [$p name]]
	}
	return $res
    }
}

proc oprs_type_va { type param string } {
    switch -glob -- [$type kind] {
	typedef - {* member} {
	    puts stderr member
	    return [oprs_type_va [$type type] $param $string]
	}
	string - array - sequence - "unsigned long" - long - float - double - int - char - enum {
	    if {[catch {$param initializer $type}]} {
		set mandat 1
	    } else {
		set mandat 0
	    }
	    return ",\"$string\",$mandat,[$type pass reference in->$string],encode_genom_[mangle $type]"
	}
	struct {
	    set res ""
	    foreach e [$type members] {
		append res [oprs_type_va [$e type] $param $string.[$e name]]
	    }
	    return $res
	}
	default {
	    puts stderr "unexpected type: [$type kind]" 
	}
    }
}

proc oprs_type_glob { type } {
    set num 0
    oprs_type $type num
}

proc oprs_service_type_glob { comp service } {
    if {[llength [$service parameters in inout]] == 0} {
	return "nil"
    } else {  
	set num 0
	set res $comp\_[$service name]_input
	foreach  p [$service parameters in inout] {
	    append res " ([$p name] [oprs_type [$p type] num])"
	    incr num
	}
	return "($res)"
    }
}

proc oprs_type { type &num} {
    upvar ${&num} num
    switch -glob -- [$type kind] {
	typedef - {* member} {
	    return [oprs_type [$type type] num]
	}
	enum {
	    return "\$[mangle $type]"
	}
	float - double {
	    return "\$float\_$num"
	}
	string {
	    if {[catch {$type length} l]} {
		return  "\$dynstring\_$num"
	    } else {
		return  "\$string$l\_$num"
	    }
	}
	int {
	    return "\$int\_$num"
	}
	char {
	    return "\$char\_$num"
	}
	long {
	    return "\$long\_$num"
	}
	"unsigned long" {
	    return "\$ulong\_$num"
	}
	sequence {
	    if {[catch {$type length} l]} {
		return "\$dynseq\_$num"
	    } else {
		return "\$seq$l\_$num"
	    }
	}
	array {
	    return "\$array[$type length]\_$num"
	}
	struct {
	    set res [mangle $type]
	    foreach e [$type members] {
		append res " ([$e name] [oprs_type [$e type] num])"
		incr num
	    }
	    return "($res)"
	}
	default {
	    puts stderr "unexpected type: [$type kind]" 
	}
    }
}


# --- isfixed --------------------------------------------------------------
#
#
proc isfixed { type } {
  set type [$type final]
  switch -- [$type kind] {
    struct - union {
      foreach e [$type members] {
        if {![isfixed $e]} { return no }
      }
      return yes
    }

    sequence {
      if {[catch {$type length}]} { return no }
      return [isfixed [$type type]]
    }

    array {
      return [isfixed [$type type]]
    }

    string {
      if {[catch {$type length}]} { return no }
      return yes
    }

    default {
      return yes
    }
  }

  error "reached unreachable"
}


# --- pubtypes -------------------------------------------------------------
#
# Return the list of all public types (recursive)
#
proc pubtypes { private } {
  set pub_types [dict create]

  foreach comp [dotgen components] {
    expandtype [$comp event] pub_types

    foreach p [$comp ports] {
      expandtype [$p datatype] pub_types
      if {$private} { expandtype [$p type] pub_types }
    }

    foreach s [$comp services] {
      foreach p [$s parameters] {
        switch [$p dir] {
          "inport" - "outport" {
            # handled above
          }
          default {
            expandtype [$p type] pub_types
          }
        }
      }
    }

    if {$private} { expandtype [$comp ids] pub_types }
  }

  return [dict values $pub_types]
}

proc expandtype { type done& } {
  upvar ${done&} done
  set name [mangle $type]
  if {[dict exists $done $name]} return

  switch -- [$type kind] {
    {array} - {sequence} - {typedef} - {struct member} - {union member} {
      expandtype [$type type] done
    }
    {struct} - {union} {
      foreach e [$type members] {
        expandtype $e done
      }
    }
  }

  if {[dict exists $done $name]} return
  dict set done $name $type
}


# --- mangle ---------------------------------------------------------------
#
#
proc mangle { type } {
  switch -- [$type kind] {
    enum - struct - union - typedef {
      return [cname [$type fullname]]
    }

    array - sequence {
      if {[catch {$type length} l]} {
        return dyn[$type kind]_[mangle [$type type]]
      } else {
        return [$type kind]${l}_[mangle [$type type]]
      }
    }

    string {
      if {[catch {$type length} l]} {
        return dynstring
      } else {
        return string${l}
      }
    }

    {struct member} - {union member} - {const} {
      return [mangle [$type type]]
    }

    default {
      return [cname [$type kind]]
    }
  }

  error "reached unreachable"
}
